# ruff: noqa: D100, D101, D102, D103, D104, D107, PLR2004
from __future__ import annotations

import grp
import logging
import os
import pwd
import socket
import stat
import subprocess
import sys
from pathlib import Path
from threading import Thread
from typing import Sequence

from ubo_app.constants import DOCKER_INSTALLATION_LOCK_FILE, USERNAME
from ubo_app.logging import add_file_handler, add_stdout_handler, get_logger
from ubo_app.system.system_manager.led import LEDManager

SOCKET_PATH = Path(os.environ.get('RUNTIME_DIRECTORY', '/run/ubo')).joinpath(
    'system_manager.sock',
)
# The order of the pixel colors - RGB or GRB.
# Some NeoPixels have red and green reversed!
# For RGBW NeoPixels, simply change the ORDER to RGBW or GRBW.

logger = get_logger('system-manager')
add_file_handler(logger, logging.DEBUG)
add_stdout_handler(logger, logging.DEBUG)
logger.setLevel(logging.DEBUG)
logger.debug('Initialising System-Manager...')


if __name__ == '__main__':
    led_manager = LEDManager()

    def run_led_command(incoming: Sequence[str]) -> Thread:
        logger.debug('---starting led new thread--', extra={'incoming': incoming})
        thread = Thread(target=led_manager.run_command, args=(incoming,))
        thread.start()
        return thread

    incoming = 'spinning_wheel 255 255 255 50 6 100'.split()
    thread = run_led_command(incoming)

    uid = pwd.getpwnam('root').pw_uid
    gid = grp.getgrnam(USERNAME).gr_gid

    SOCKET_PATH.unlink(missing_ok=True)

    logger.info('System Manager opening socket...')
    server = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
    server.bind(SOCKET_PATH.as_posix())
    permission = (
        stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH | stat.S_IWGRP | stat.S_IWUSR
    )
    SOCKET_PATH.chmod(permission)

    os.chown(SOCKET_PATH, uid, gid)

    logger.info('System Manager Listening...')
    while True:
        try:
            datagram = server.recv(1024)
            if not datagram:
                break
            else:
                incoming_str = datagram.decode('utf-8')
                logger.debug('Received:', extra={'incoming_str': incoming_str})
                header, *incoming = incoming_str.split()
                if header == 'led':
                    led_manager.stop()
                    thread.join()
                    thread = run_led_command(incoming)
                elif header == 'docker':
                    # Run the install_docker.sh script
                    def install_docker(command: str) -> None:
                        Path.touch(
                            DOCKER_INSTALLATION_LOCK_FILE,
                            mode=0o666,
                            exist_ok=True,
                        )
                        try:
                            if command == 'install':
                                subprocess.run(
                                    Path(__file__)  # noqa: S603
                                    .parent.parent.joinpath('install_docker.sh')
                                    .as_posix(),
                                    env={'USERNAME': USERNAME},
                                    check=False,
                                )
                            elif command == 'start':
                                subprocess.run(
                                    [  # noqa: S603
                                        '/usr/bin/env',
                                        'systemctl',
                                        'start',
                                        'docker.socket',
                                    ],
                                    check=False,
                                )
                                subprocess.run(
                                    [  # noqa: S603
                                        '/usr/bin/env',
                                        'systemctl',
                                        'start',
                                        'docker.service',
                                    ],
                                    check=False,
                                )
                            elif command == 'stop':
                                subprocess.run(
                                    [  # noqa: S603
                                        '/usr/bin/env',
                                        'systemctl',
                                        'stop',
                                        'docker.socket',
                                    ],
                                    check=False,
                                )
                                subprocess.run(
                                    [  # noqa: S603
                                        '/usr/bin/env',
                                        'systemctl',
                                        'stop',
                                        'docker.service',
                                    ],
                                    check=False,
                                )
                        finally:
                            DOCKER_INSTALLATION_LOCK_FILE.unlink(missing_ok=True)

                    thread = Thread(target=install_docker, args=(incoming[0],))
                    thread.start()
        except KeyboardInterrupt:
            logger.debug('Interrupted')
            server.close()
            try:
                sys.exit(0)
            except SystemExit:
                os._exit(0)
    logger.info('Shutting down...')
    server.close()
    SOCKET_PATH.unlink()



"""Application logic."""

from __future__ import annotations

import atexit
import subprocess
import weakref
from typing import TYPE_CHECKING, Sequence

from debouncer import DebounceOptions, debounce
from kivy.clock import mainthread
from redux import AutorunOptions, FinishAction, FinishEvent

from ubo_app.store import autorun, dispatch, subscribe_event
from ubo_app.store.main import PowerOffEvent
from ubo_app.store.services.notifications import Chime
from ubo_app.store.services.sound import SoundPlayChimeAction
from ubo_app.store.update_manager import (
    UpdateManagerCheckEvent,
    UpdateManagerSetStatusAction,
    UpdateManagerUpdateEvent,
    UpdateStatus,
)
from ubo_app.store.update_manager.reducer import ABOUT_MENU_PATH
from ubo_app.store.update_manager.utils import check_version, update
from ubo_app.utils.async_ import create_task
from ubo_app.utils.hardware import (
    IS_RPI,
    initialize_board,
    turn_off_screen,
    turn_on_screen,
)

if TYPE_CHECKING:
    from ubo_app.menu import MenuApp


def power_off() -> None:
    """Power off the device."""
    dispatch(SoundPlayChimeAction(name=Chime.FAILURE), FinishAction())
    if IS_RPI:
        subprocess.run(['/usr/bin/env', 'systemctl', 'poweroff', '-i'], check=True)  # noqa: S603


def setup(app: MenuApp) -> None:
    """Set up the application."""
    turn_on_screen()
    initialize_board()

    subscribe_event(PowerOffEvent, power_off)

    app_ref = weakref.ref(app)

    @mainthread
    def stop_app() -> None:
        app = app_ref()
        if app is not None:
            app.stop()

    subscribe_event(FinishEvent, stop_app)
    subscribe_event(UpdateManagerUpdateEvent, update)
    subscribe_event(UpdateManagerCheckEvent, check_version)

    @debounce(
        wait=10,
        options=DebounceOptions(leading=True, trailing=False, time_window=10),
    )
    async def request_check_version() -> None:
        dispatch(UpdateManagerSetStatusAction(status=UpdateStatus.CHECKING))

    @autorun(lambda state: state.main.path, options=AutorunOptions(keep_ref=False))
    def _(
        path: Sequence[str],
        previous_path: Sequence[str],
    ) -> None:
        if path != previous_path and path[:3] == ABOUT_MENU_PATH:
            create_task(request_check_version())

    create_task(request_check_version())

    atexit.register(turn_off_screen)



# ruff: noqa: D100, D101, D102, D103, D104, D105, D107, N999
from __future__ import annotations

import pathlib
from typing import Sequence

from kivy.clock import mainthread
from kivy.lang.builder import Builder
from kivy.properties import BooleanProperty
from ubo_gui.constants import SUCCESS_COLOR
from ubo_gui.menu.types import ActionItem, Item
from ubo_gui.page import PageWidget
from wifi_manager import add_wireless_connection

from ubo_app.logging import logger
from ubo_app.store import dispatch, subscribe_event
from ubo_app.store.services.camera import CameraStartViewfinderAction
from ubo_app.store.services.notifications import (
    Chime,
    Notification,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.store.services.sound import SoundPlayChimeAction
from ubo_app.store.services.wifi import (
    WiFiCreateEvent,
    WiFiType,
    WiFiUpdateRequestAction,
)
from ubo_app.utils.async_ import create_task

# Regular expression pattern
# WIFI:S:<SSID>;T:<WEP|WPA|blank>;P:<PASSWORD>;H:<true|false|blank>;;
barcode_pattern = (
    r'^WIFI:S:(?P<SSID>[^;]*);(?:T:(?P<Type>(?i:WEP|WPA|WPA2|nopass));)'
    r'?(?:P:(?P<Password>[^;]*);)?(?:H:(?P<Hidden>(?i:true|false));)?$'
)


class CreateWirelessConnectionPage(PageWidget):
    creating = BooleanProperty(defaultvalue=False)

    def __init__(
        self: CreateWirelessConnectionPage,
        items: Sequence[Item] | None = None,
        *args: object,
        **kwargs: object,
    ) -> None:
        super().__init__(*args, **kwargs, items=items)
        self.unsubscribe = subscribe_event(
            WiFiCreateEvent,
            self.create_wireless_connection,
        )
        dispatch(SoundPlayChimeAction(name='scan'))
        self.bind(on_close=lambda *_: self.unsubscribe())

    def create_wireless_connection(
        self: CreateWirelessConnectionPage,
        event: WiFiCreateEvent,
    ) -> None:
        connection = event.connection
        ssid = connection.ssid
        password = connection.password

        if not password:
            logger.warn('Password is required')
            return

        async def act() -> None:
            await add_wireless_connection(
                ssid=ssid,
                password=password,
                type=connection.type or WiFiType.nopass,
                hidden=connection.hidden,
            )

            logger.info(
                'Wireless connection created',
                extra={
                    'ssid': ssid,
                    'type': connection.type,
                    'hidden': connection.hidden,
                },
            )

            dispatch(
                WiFiUpdateRequestAction(reset=True),
                NotificationsAddAction(
                    notification=Notification(
                        title=f'"{ssid}" Added',
                        content=f"""WiFi connection with ssid "{
                        ssid}" was added successfully""",
                        display_type=NotificationDisplayType.FLASH,
                        color=SUCCESS_COLOR,
                        icon='wifi_add',
                        chime=Chime.ADD,
                    ),
                ),
            )
            mainthread(self.dispatch)('on_close')

        self.creating = True
        create_task(act())

    def get_item(self: CreateWirelessConnectionPage, index: int) -> ActionItem | None:
        if index == 2:  # noqa: PLR2004
            return ActionItem(
                label='start',
                is_short=True,
                icon='camera',
                action=lambda: dispatch(
                    CameraStartViewfinderAction(barcode_pattern=barcode_pattern),
                ),
            )
        return super().get_item(index)


Builder.load_file(
    pathlib.Path(__file__)
    .parent.joinpath('create_wireless_connection.kv')
    .resolve()
    .as_posix(),
)



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
import math

WIFI_STATE_ICON_ID = 'wifi:state'
WIFI_STATE_ICON_PRIORITY = -12
SIGNAL_ICONS = [
    'signal_wifi_0_bar',
    'network_wifi_1_bar',
    'network_wifi_2_bar',
    'network_wifi_3_bar',
    'signal_wifi_4_bar',
]


def get_signal_icon(strength: float) -> str:
    return SIGNAL_ICONS[math.floor(strength / 100 * 4.999)]



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

import re
import time
from typing import TYPE_CHECKING

import headless_kivy_pi.config
import numpy as np
from kivy.clock import Clock

from ubo_app.logging import logger
from ubo_app.store import dispatch, subscribe_event
from ubo_app.store.services.camera import (
    CameraBarcodeAction,
    CameraStartViewfinderEvent,
    CameraStopViewfinderEvent,
)
from ubo_app.utils import IS_RPI

if TYPE_CHECKING:
    from numpy._typing import NDArray

THROTTL_TIME = 0.5


def resize_image(
    image: NDArray[np.uint8],
    new_size: tuple[int, int] = (
        headless_kivy_pi.config.width(),
        headless_kivy_pi.config.height(),
    ),
) -> NDArray[np.uint8]:
    scale_x = image.shape[1] / new_size[1]
    scale_y = image.shape[0] / new_size[0]

    # Use slicing to downsample the image
    resized = image[:: int(scale_y), :: int(scale_x)]

    # Handle any rounding issues by trimming the excess
    return resized[: new_size[0], : new_size[1]]


def check_codes(regex: re.Pattern, codes: list[str], last_match: float) -> float:
    if time.time() - last_match < THROTTL_TIME:
        return last_match

    for code in codes:
        logger.info(
            'Read barcode, decoded value',
            extra={'decoded_value': code},
        )
        match = regex.match(code)
        if match:
            logger.info(
                'Pattern match',
                extra={
                    'pattern': regex.pattern,
                    'match': match.groupdict(),
                    'decoded_value': code,
                },
            )
            dispatch(CameraBarcodeAction(code=code, match=match.groupdict()))

    return time.time()


def init_service() -> None:
    if not IS_RPI:
        subscribe_event(
            CameraStartViewfinderEvent,
            lambda event: check_codes(
                re.compile(event.barcode_pattern or ''),
                ['WIFI:S:SSID;T:WPA;P:password;;'],
                0,
            ),
        )
        return

    from picamera2 import Picamera2  # pyright: ignore [reportMissingImports]
    from pyzbar.pyzbar import decode

    picam2 = Picamera2()
    preview_config = picam2.create_still_configuration(
        {
            'format': 'RGB888',
            'size': (
                headless_kivy_pi.config.width() * 2,
                headless_kivy_pi.config.height() * 2,
            ),
        },
    )
    picam2.configure(preview_config)
    picam2.set_controls({'AwbEnable': True})

    picam2.start()

    def start_camera_viewfinder(start_event: CameraStartViewfinderEvent) -> None:
        regex_pattern = start_event.barcode_pattern
        regex = re.compile(regex_pattern) if regex_pattern is not None else None
        last_match = 0

        display = headless_kivy_pi.config._display  # noqa: SLF001
        if not display:
            return

        def feed_viewfinder(_: object) -> None:
            display = headless_kivy_pi.config._display  # noqa: SLF001
            if not display:
                return
            data = picam2.capture_array('main')

            barcodes = decode(data)
            if len(barcodes) > 0 and regex is not None:
                nonlocal last_match
                last_match = check_codes(
                    regex,
                    [barcode.data.decode() for barcode in barcodes],
                    last_match,
                )

            data = resize_image(data)
            data = np.rot90(data, 2)

            # Mirror the image
            data = data[:, ::-1, :3].astype(np.uint16)
            color = (
                ((data[:, :, 0] & 0xF8) << 8)
                | ((data[:, :, 1] & 0xFC) << 3)
                | (data[:, :, 2] >> 3)
            )
            data_bytes = bytes(
                np.dstack(((color >> 8) & 0xFF, color & 0xFF)).flatten().tolist(),
            )

            display._block(  # noqa: SLF001
                0,
                0,
                headless_kivy_pi.config.width() - 1,
                headless_kivy_pi.config.height() - 1,
                data_bytes,
            )

        feed_viewfinder_scheduler = Clock.schedule_interval(feed_viewfinder, 0.03)

        headless_kivy_pi.config.pause()

        def handle_stop_viewfinder() -> None:
            feed_viewfinder_scheduler.cancel()
            headless_kivy_pi.config.resume()
            cancel_subscription()

        cancel_subscription = subscribe_event(
            CameraStopViewfinderEvent,
            handle_stop_viewfinder,
        )

    subscribe_event(CameraStartViewfinderEvent, start_camera_viewfinder)



# pyright: reportMissingImports=false,reportMissingModuleSource=false
"""Module for managing audio playback and recording."""
from __future__ import annotations

import contextlib
import math
import time
import wave

from ubo_app.logging import logger
from ubo_app.utils import IS_RPI

if not IS_RPI:
    import sys

    from ubo_app.utils.fake import Fake

    sys.modules['alsaaudio'] = Fake()
    sys.modules['pulsectl'] = Fake()
    sys.modules['pyaudio'] = Fake()

import alsaaudio
import pulsectl
import pyaudio

CHUNK_SIZE = 1024


def _linear_to_logarithmic(volume_linear: float) -> int:
    """Convert a linear volume to a logarithmic volume.

    Assuming volume_linear is between 0 and 1
    Convert it to a logarithmic scale.
    """
    if volume_linear == 0:
        return 0
    return round(100 * math.log(volume_linear * 500) / math.log(500))


class AudioManager:
    """Class for managing audio playback and recording."""

    stream: pyaudio.PyAudio.Stream | None = None

    def __init__(self: AudioManager) -> None:
        """Initialize the audio manager."""
        # create an audio object
        self.pyaudio = pyaudio.PyAudio()
        self.is_playing = False
        self.should_stop = False

        self.cardindex = None
        try:
            cards = alsaaudio.cards()
            self.cardindex = cards.index(
                next(card for card in cards if 'wm8960' in card),
            )
            try:
                with pulsectl.Pulse('set-default-sink') as pulse:
                    for sink in pulse.sink_list():
                        if str(sink.proplist['alsa.card']) == str(self.cardindex):
                            pulse.sink_default_set(sink)
                            break
            except pulsectl.PulseError:
                logger.error('Not able to connect to pulseaudio')
        except StopIteration:
            logger.error('No audio card found')

    def find_respeaker_index(self: AudioManager) -> int:
        """Find the index of the ReSpeaker device."""
        for index in range(self.pyaudio.get_device_count()):
            info = self.pyaudio.get_device_info_by_index(index)
            if not isinstance(info['name'], (int, float)) and 'wm8960' in info['name']:
                logger.debug(f'ReSpeaker found at index: {index}')
                logger.debug(f'Device Info: {info}')
                return index
        msg = 'ReSpeaker for default device not found'
        raise ValueError(msg)

    def __del__(self: AudioManager) -> None:
        """Clean up the audio manager."""
        self.close_stream()
        self.pyaudio.terminate()

    def close_stream(self: AudioManager) -> None:
        """Clean up the audio manager."""
        if self.stream:
            self.should_stop = True
            while self.is_playing:
                time.sleep(0.05)
            self.should_stop = False
            with contextlib.suppress(Exception):
                self.stream.close()
            self.stream = None

    def play(self: AudioManager, filename: str) -> None:
        """Play a waveform audio file.

        Parameters
        ----------
        filename : str
            Path to wav file

        """
        # open the file for reading.
        self.close_stream()

        logger.info('Opening audio file for playback', extra={'filename_': filename})
        try:
            with wave.open(filename, 'rb') as wf:
                self.is_playing = True
                stream = self.stream = self.pyaudio.open(
                    format=self.pyaudio.get_format_from_width(wf.getsampwidth()),
                    channels=wf.getnchannels(),
                    rate=wf.getframerate(),
                    output=True,
                    output_device_index=self.find_respeaker_index(),
                )
                data = wf.readframes(CHUNK_SIZE)
                while data and not self.should_stop and stream.is_active():
                    stream.write(data)
                    data = wf.readframes(CHUNK_SIZE)
        except Exception as exception:  # noqa: BLE001
            logger.error(
                'Something went wrong while playing an audio file',
                exc_info=exception,
            )
        finally:
            self.is_playing = False
            self.close_stream()

    def set_playback_mute(self: AudioManager, *, mute: bool = False) -> None:
        """Set the playback mute of the audio output.

        Parameters
        ----------
        mute : bool
            Mute to set

        """
        try:
            # Assume pulseaudio is installed
            mixer = alsaaudio.Mixer(control='Master')
            mixer.setmute(1 if mute else 0)
        except alsaaudio.ALSAAudioError:
            # Seems like pulseaudio is not installed, so we directly use device mixers
            if self.cardindex is None:
                return

            mixer = alsaaudio.Mixer(
                control='Right Output Mixer PCM',
                cardindex=self.cardindex,
            )
            mixer.setmute(0)
            mixer = alsaaudio.Mixer(
                control='Left Output Mixer PCM',
                cardindex=self.cardindex,
            )
            mixer.setmute(0)

    def set_playback_volume(self: AudioManager, volume: float = 0.8) -> None:
        """Set the playback volume of the audio output.

        Parameters
        ----------
        volume : float
            Volume to set, a float between 0 and 1

        """
        if volume < 0 or volume > 1:
            msg = 'Volume must be between 0 and 1'
            raise ValueError(msg)
        try:
            # Assume pulseaudio is installed
            mixer = alsaaudio.Mixer(control='Master')
            mixer.setvolume(round(volume * 100))
        except alsaaudio.ALSAAudioError:
            # Seems like pulseaudio is not installed, so we directly use device mixers
            if self.cardindex is None:
                return

            mixer = alsaaudio.Mixer(control='Speaker', cardindex=self.cardindex)
            mixer.setvolume(
                _linear_to_logarithmic(volume),
                alsaaudio.MIXER_CHANNEL_ALL,
                alsaaudio.PCM_PLAYBACK,
            )
            mixer = alsaaudio.Mixer(control='Playback', cardindex=self.cardindex)
            mixer.setvolume(
                100,
                alsaaudio.MIXER_CHANNEL_ALL,
                alsaaudio.PCM_PLAYBACK,
            )

    def set_capture_volume(self: AudioManager, volume: float = 0.8) -> None:
        """Set the capture volume of the audio output.

        Parameters
        ----------
        volume : float
            Volume to set, a float between 0 and 1

        """
        if volume < 0 or volume > 1:
            msg = 'Volume must be between 0 and 1'
            raise ValueError(msg)
        try:
            # Assume pulseaudio is installed
            mixer = alsaaudio.Mixer(control='Capture')
            mixer.setrec(round(volume * 100))
        except alsaaudio.ALSAAudioError:
            # Seems like pulseaudio is not installed, so we directly use device mixers
            if self.cardindex is None:
                return

            mixer = alsaaudio.Mixer(control='Capture', cardindex=self.cardindex)
            mixer.setvolume(
                _linear_to_logarithmic(volume),
                alsaaudio.MIXER_CHANNEL_ALL,
                alsaaudio.PCM_CAPTURE,
            )



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from setup import init_service

from ubo_app.load_services import register_service

register_service(
    service_id='ethernet',
    label='Ethernet',
    init=init_service,
)



# pyright: reportMissingImports=false
# ruff: noqa: D100, D101, D102, D103, D104, D105, D107
from __future__ import annotations

from typing import Callable

from ubo_app.utils import IS_RPI
from ubo_app.utils.bus_provider import get_system_bus

if not IS_RPI:
    import sys

    from ubo_app.utils.fake import Fake

    sys.modules['sdbus'] = Fake()

from sdbus import DbusInterfaceCommonAsync, dbus_property_async


class SystemdUnitInterface(  # pyright: ignore [reportGeneralTypeIssues]
    DbusInterfaceCommonAsync,
    interface_name='org.freedesktop.systemd1.Unit',  # pyright: ignore [reportCallIssue]
):
    @dbus_property_async(property_signature='s')
    def active_state(self: SystemdUnitInterface) -> str:
        raise NotImplementedError


def to_dbus_string(string: str) -> str:
    """Convert a string to a D-Bus object path."""
    result = []
    for char in string:
        if char.isalnum() or char == '_':
            result.append(char)
        else:
            result.append('_%02x' % ord(char))
    return ''.join(result)


async def monitor_unit(unit_name: str, callback: Callable[[str], None]) -> None:
    bus = get_system_bus()
    system_service = SystemdUnitInterface.new_proxy(
        bus=bus,
        service_name='org.freedesktop.systemd1',
        object_path=f'/org/freedesktop/systemd1/unit/{to_dbus_string(unit_name)}',
    )

    async for _ in system_service.properties_changed:
        active_state = await system_service.active_state
        callback(active_state)



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

from typing import TypeAlias

from immutable import Immutable
from redux import BaseAction, BaseEvent

Color: TypeAlias = (
    tuple[float | int, float | int, float | int]
    | tuple[float | int, float | int, float | int, float | int]
)


class RgbRingAction(BaseAction):
    ...


class RgbRingEvent(BaseEvent):
    ...


class RgbRingSetIsConnectedAction(RgbRingAction):
    is_connected: bool | None = None


class RgbRingSetIsBusyAction(RgbRingAction):
    is_busy: bool | None = None


class RgbRingCommandAction(RgbRingAction):
    ...


class RgbRingWaitableCommandAction(RgbRingCommandAction):
    wait: int = 100


class RgbRingColorfulCommandAction(RgbRingCommandAction):
    color: Color = (255, 255, 255)


class RgbRingSetEnabledAction(RgbRingCommandAction):
    enabled: bool = True


class RgbRingSetAllAction(RgbRingColorfulCommandAction):
    ...


class RgbRingSetBrightnessAction(RgbRingCommandAction):
    brightness: float = 0.5


class RgbRingBlankAction(RgbRingCommandAction):
    ...


class RgbRingRainbowAction(RgbRingWaitableCommandAction):
    rounds: int


class RgbRingProgressWheelStepAction(RgbRingColorfulCommandAction):
    ...


class RgbRingPulseAction(RgbRingWaitableCommandAction, RgbRingColorfulCommandAction):
    repetitions: int = 5


class RgbRingBlinkAction(RgbRingWaitableCommandAction, RgbRingColorfulCommandAction):
    repetitions: int = 5


class RgbRingSpinningWheelAction(
    RgbRingWaitableCommandAction,
    RgbRingColorfulCommandAction,
):
    length: int = 5
    repetitions: int = 5


class RgbRingProgressWheelAction(RgbRingColorfulCommandAction):
    percentage: float = 0.5


class RgbRingFillUptoAction(RgbRingWaitableCommandAction, RgbRingColorfulCommandAction):
    percentage: float = 0.5


class RgbRingFillDownfromAction(
    RgbRingWaitableCommandAction,
    RgbRingColorfulCommandAction,
):
    percentage: float = 0.5


class RgbRingCommandEvent(RgbRingEvent):
    command: list[str]


class RgbRingState(Immutable):
    is_connected: bool
    is_busy: bool



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from setup import init_service

from ubo_app.load_services import register_service

register_service(
    service_id='keyboard',
    label='Keyboard',
    init=init_service,
)



# pyright: reportMissingModuleSource=false
# ruff: noqa: D100, D101, D102, D103, D104, D107
from ubo_app.utils import IS_RPI


def initialize_board() -> None:
    if not IS_RPI:
        return

    from RPi import GPIO

    GPIO.setmode(GPIO.BCM)
    GPIO.setup(17, GPIO.IN, pull_up_down=GPIO.PUD_DOWN)


def turn_off_screen() -> None:
    if not IS_RPI:
        return
    import headless_kivy_pi.config
    from headless_kivy_pi.constants import BYTES_PER_PIXEL
    from RPi import GPIO

    GPIO.setup(26, GPIO.OUT)
    GPIO.output(26, GPIO.LOW)

    display = headless_kivy_pi.config._display  # noqa: SLF001
    if not display:
        return
    data = (
        [0]
        * headless_kivy_pi.config.width()
        * headless_kivy_pi.config.height()
        * BYTES_PER_PIXEL
    )
    display._block(  # noqa: SLF001
        0,
        0,
        headless_kivy_pi.config.width() - 1,
        headless_kivy_pi.config.height() - 1,
        data,
    )


def turn_on_screen() -> None:
    if not IS_RPI:
        return

    from RPi import GPIO

    GPIO.setmode(GPIO.BCM)
    GPIO.setup(26, GPIO.OUT)
    GPIO.output(26, GPIO.HIGH)



# ruff: noqa: D100, D101, D102, D103, D104, D105, D107
from __future__ import annotations

import asyncio
import importlib
import importlib.abc
import importlib.util
import os
import sys
import threading
import traceback
import uuid
from importlib.machinery import PathFinder, SourceFileLoader
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, ClassVar, Coroutine, Sequence, cast

from redux import CombineReducerRegisterAction, FinishEvent, ReducerType

from ubo_app.constants import DEBUG_MODE, SERVICES_PATH
from ubo_app.logging import logger

if TYPE_CHECKING:
    from importlib.machinery import ModuleSpec
    from types import ModuleType

ROOT_PATH = Path(__file__).parent
REGISTERED_PATHS: dict[Path, UboServiceThread] = {}


# Customized module finder and module loader for ubo services to avoid mistakenly
# loading conflicting names in different services.
# This is a temporary hack until each service runs in its own process.
class UboServiceModuleLoader(SourceFileLoader):
    cache: ClassVar[dict[str, ModuleType]] = {}

    @property
    def cache_id(self: UboServiceModuleLoader) -> str:
        return f'{self.path}:{self.name}'

    def create_module(
        self: UboServiceModuleLoader,
        spec: ModuleSpec,
    ) -> ModuleType | None:
        _ = spec
        if self.cache_id in UboServiceModuleLoader.cache:
            return UboServiceModuleLoader.cache[self.cache_id]
        return None

    def exec_module(self: UboServiceModuleLoader, module: ModuleType) -> None:
        if self.cache_id in UboServiceModuleLoader.cache:
            return
        super().exec_module(module)
        UboServiceModuleLoader.cache[self.cache_id] = module

    def get_filename(self: UboServiceModuleLoader, name: str | None = None) -> str:
        return super().get_filename(name.split(':')[-1] if name else name)


class UboServiceLoopLoader(importlib.abc.Loader):
    def __init__(self: UboServiceLoopLoader, service: UboServiceThread) -> None:
        self.service = service

    def exec_module(self: UboServiceLoopLoader, module: ModuleType) -> None:
        cast(Any, module).current_loop = lambda: self.service.loop
        cast(Any, module)._create_task = (  # noqa: SLF001
            lambda task: self.service.loop.call_soon_threadsafe(
                self.service.loop.create_task,
                task,
            )
        )
        cast(Any, module)._run_in_executor = (  # noqa: SLF001
            lambda executor, task, *args: self.service.loop.run_in_executor(
                executor,
                task,
                *args,
            )
        )

    def __repr__(self: UboServiceLoopLoader) -> str:
        return f'{self.service.path}'


class UboServiceFinder(importlib.abc.MetaPathFinder):
    def find_spec(
        self: UboServiceFinder,
        fullname: str,
        _: Sequence[str] | None,
        target: ModuleType | None = None,
    ) -> ModuleSpec | None:
        stack = traceback.extract_stack()
        matching_path = next(
            (
                registered_path
                for stack_path in stack[-2::-1]
                for registered_path in (*REGISTERED_PATHS.keys(), Path('/'))
                if stack_path.filename.startswith(registered_path.as_posix())
            ),
            None,
        )

        if matching_path in REGISTERED_PATHS:
            service = REGISTERED_PATHS[matching_path]
            module_name = f'{service.service_uid}:{fullname}'

            if fullname == 'ubo_app.utils.loop':
                return importlib.util.spec_from_loader(
                    module_name,
                    UboServiceLoopLoader(service),
                )

            spec = PathFinder.find_spec(
                fullname,
                [matching_path.as_posix()],
                target,
            )
            if spec and spec.origin:
                spec.name = module_name
                spec.loader = UboServiceModuleLoader(fullname, spec.origin)
            return spec
        return None


sys.meta_path.insert(0, UboServiceFinder())


class UboServiceThread(threading.Thread):
    def __init__(
        self: UboServiceThread,
        path: Path,
        white_list: Sequence[str] | None = None,
    ) -> None:
        name = path.name
        super().__init__()
        self.service_uid = f'{uuid.uuid4().hex}:{name}'
        self.name = name
        self.label = '<NOT SET>'
        self.service_id = '-not-set-'

        self.path = path
        self.white_list = white_list

        self.loop = asyncio.new_event_loop()
        self.module = None
        if DEBUG_MODE:
            self.loop.set_debug(enabled=True)

    def run(self: UboServiceThread) -> None:
        from ubo_app import store

        try:
            if self.path.exists():
                module_name = f'{self.service_uid}:ubo_handle'
                self.spec = PathFinder.find_spec(
                    'ubo_handle',
                    [self.path.as_posix()],
                    None,
                )
                if not self.spec or not self.spec.origin:
                    return
                self.spec.name = module_name
                self.spec.loader = UboServiceModuleLoader(
                    'ubo_handle',
                    self.spec.origin,
                )
                self.module = importlib.util.module_from_spec(self.spec)

        except Exception as exception:  # noqa: BLE001
            logger.error(f'Error loading "{self.path}"', exc_info=exception)

        asyncio.set_event_loop(self.loop)
        if self.module and self.spec and self.spec.loader:
            try:
                self.spec.loader.exec_module(self.module)
            except Exception as exception:  # noqa: BLE001
                logger.error(f'Error loading "{self.path}"', exc_info=exception)
                return

        store.subscribe_event(FinishEvent, self.stop)
        self.loop.run_forever()

    def stop(self: UboServiceThread) -> None:
        self.loop.call_soon_threadsafe(self.loop.stop)

    def __repr__(self: UboServiceThread) -> str:
        return f'<UboServiceThread id={self.name} label={self.label}>'


def register_service(
    service_id: str,
    label: str,
    reducer: ReducerType | None = None,
    init: Callable[[], None | Coroutine] | None = None,
) -> None:
    from ubo_app import store

    thread = threading.current_thread()
    if not isinstance(thread, UboServiceThread):
        logger.error(
            'Service registration outside of service thread',
            extra={
                'service_id': service_id,
                'label': label,
            },
        )
        return

    if service_id in os.environ.get('UBO_DISABLED_SERVICES', '').split(','):
        logger.info(
            'Skipping disabled ubo service',
            extra={
                'service_id': service_id,
                'label': label,
                'disabled_services': os.environ.get('UBO_DISABLED_SERVICES'),
            },
        )
        thread.stop()
        return

    if thread.white_list and service_id not in thread.white_list:
        logger.info(
            'Service is not in services white list',
            extra={
                'service_id': service_id,
                'label': label,
                'white_list': thread.white_list,
            },
        )
        thread.stop()
        return

    thread.label = label
    thread.service_id = service_id

    logger.info(
        'Registering ubo serivce',
        extra={
            'service_id': service_id,
            'label': label,
            'has_reducer': reducer is not None,
        },
    )

    if reducer is not None:
        store.dispatch(
            CombineReducerRegisterAction(
                _id=store.root_reducer_id,
                key=service_id,
                reducer=reducer,
            ),
        )

    if init is not None:
        result = init()
        if asyncio.iscoroutine(result):
            thread.loop.create_task(result)


def load_services(service_ids: Sequence[str] | None = None) -> None:
    for services_directory_path in [
        ROOT_PATH.joinpath('services').as_posix(),
        *SERVICES_PATH,
    ]:
        if Path(services_directory_path).is_dir():
            for service_path in Path(services_directory_path).iterdir():
                if not service_path.is_dir():
                    continue
                current_path = Path().absolute()
                os.chdir(service_path.as_posix())

                service = UboServiceThread(service_path, white_list=service_ids)
                REGISTERED_PATHS[service_path] = service
                service.start()

                os.chdir(current_path)



"""Implement `setup_service` function to set up and enable systemd service."""
from __future__ import annotations

import grp
import os
import pwd
import subprocess
import time
import warnings
from pathlib import Path
from typing import Literal, TypedDict

from ubo_app.constants import INSTALLATION_PATH, USERNAME
from ubo_app.logging import logger

RETRIES = 5


class Service(TypedDict):
    """System service metadata."""

    name: str
    template: str
    scope: Literal['system', 'user']


services: list[Service] = [
    {
        'name': 'ubo-system',
        'template': 'system',
        'scope': 'system',
    },
    {
        'name': 'ubo-update',
        'template': 'update',
        'scope': 'system',
    },
    {
        'name': 'ubo-app',
        'template': 'app',
        'scope': 'user',
    },
]

uid = pwd.getpwnam(USERNAME).pw_uid
gid = grp.getgrnam(USERNAME).gr_gid


def create_user_service_directory() -> None:
    """Create the user service file."""
    path = Path(f'/home/{USERNAME}/.config/systemd/user')
    path.mkdir(parents=True, exist_ok=True)
    while path != Path(f'/home/{USERNAME}'):
        os.chown(path, uid, gid)
        path = path.parent


def create_service_file(service: Service) -> None:
    """Create the service file."""
    if service['scope'] == 'user':
        service_file_path = (
            f'/home/{USERNAME}/.config/systemd/user/{service["name"]}.service'
        )
    elif service['scope'] == 'system':
        service_file_path = f'/etc/systemd/system/{service["name"]}.service'
    else:
        msg = f"Service '{service['name']}' has an invalid scope: {service['scope']}"
        logger.error(msg)
        raise ValueError(msg)

    template = (
        Path(__file__)
        .parent.joinpath(f'services/{service["template"]}.service')
        .open()
        .read()
    )

    content = template.replace(
        '{{INSTALLATION_PATH}}',
        INSTALLATION_PATH,
    ).replace(
        '{{USERNAME}}',
        USERNAME,
    )

    # Write the service content to the file
    with Path(service_file_path).open('w') as file:
        file.write(content)
        if service['scope'] == 'user':
            os.chown(service_file_path, uid, gid)


def reload_daemon() -> None:
    """Reload the systemd daemon for the user and system services."""
    logger.info('Waiting for the user services to come up...')
    for i in range(RETRIES):
        time.sleep(1)
        print('.', end='', flush=True)  # noqa: T201
        try:
            subprocess.run(
                [  # noqa: S603
                    '/usr/bin/env',
                    'sudo',
                    f'XDG_RUNTIME_DIR=/run/user/{uid}',
                    '-u',
                    USERNAME,
                    'systemctl',
                    '--user',
                    'daemon-reload',
                ],
                check=True,
            )
        except subprocess.CalledProcessError as exception:
            if i < RETRIES - 1:
                logger.error(
                    'Failed to reload user services, retrying...',
                    exc_info=exception,
                )
                continue
        else:
            break
    else:
        msg = f'Failed to reload user services after {RETRIES} times, giving up!'
        logger.error(msg)
        warnings.warn(msg, stacklevel=2)
    print(flush=True)  # noqa: T201
    subprocess.run(
        ['/usr/bin/env', 'systemctl', 'daemon-reload'],  # noqa: S603
        check=True,
    )


def enable_services() -> None:
    """Enable the services to start on boot."""
    for service in services:
        # Enable the service to start on boot
        if service['scope'] == 'user':
            subprocess.run(
                [  # noqa: S603
                    '/usr/bin/env',
                    'sudo',
                    f'XDG_RUNTIME_DIR=/run/user/{uid}',
                    '-u',
                    USERNAME,
                    'systemctl',
                    '--user',
                    'enable',
                    service['name'],
                ],
                check=True,
            )
        elif service['scope'] == 'system':
            subprocess.run(
                ['/usr/bin/env', 'systemctl', 'enable', service['name']],  # noqa: S603
                check=True,
            )

        logger.info(
            f"Service '{service['name']}' has been created and enabled.",
        )


def setup_polkit() -> None:
    """Create the polkit rules file."""
    with Path('/etc/polkit-1/rules.d/50-ubo.rules').open('w') as file:
        file.write(
            Path(__file__)
            .parent.joinpath('polkit.rules')
            .open()
            .read()
            .replace('{{INSTALLATION_PATH}}', INSTALLATION_PATH)
            .replace('{{USERNAME}}', USERNAME),
        )


def install_docker() -> None:
    """Install docker."""
    logger.info('Installing docker...')
    for i in range(RETRIES):
        time.sleep(1)
        print('.', end='', flush=True)  # noqa: T201
        try:
            subprocess.run(
                [Path(__file__).parent.joinpath('install_docker.sh').as_posix()],  # noqa: S603
                env={'USERNAME': USERNAME},
                check=True,
            )
        except subprocess.CalledProcessError as exception:
            if i < RETRIES - 1:
                logger.error(
                    'Failed to install docker, retrying...',
                    exc_info=exception,
                )
                continue
        else:
            break
    else:
        logger.error(f'Failed to installed docker {RETRIES} times, giving up!')
        return
    print(flush=True)  # noqa: T201


def bootstrap(*, with_docker: bool = False, for_packer: bool = False) -> None:
    """Create the service files and enable the services."""
    # Ensure we have the required permissions
    if os.geteuid() != 0:
        logger.error('This script needs to be run with root privileges.')
        return

    create_user_service_directory()

    for service in services:
        create_service_file(service)

    if for_packer:
        Path('/var/lib/systemd/linger').mkdir(exist_ok=True, parents=True)
        Path(f'/var/lib/systemd/linger/{USERNAME}').touch(mode=0o644, exist_ok=True)
    else:
        subprocess.run(
            ['/usr/bin/env', 'loginctl', 'enable-linger', USERNAME],  # noqa: S603
            check=True,
        )

    reload_daemon()
    enable_services()

    setup_polkit()

    if with_docker:
        install_docker()

    subprocess.run(
        [Path(__file__).parent.joinpath('install_wm8960.sh').as_posix()],  # noqa: S603
        check=True,
    )
